from anyio import aopen, open_cancel_scope

import logging

def combine_dict(*d):
    """
    Returns a dict with all keys+values of all dict arguments.
    The first found value wins.
    
    This recurses if values are dicts.
    """
    res = {}
    keys = {}
    if len(d) <= 1:
        return d
    for kv in d:
        for k,v in kv.items():
            if k not in keys:
                keys[k] = []
            keys[k].append(v)
    for k,v in keys.items():
        if len(v) == 1:
            res[k] = v[0]
        elif not isinstance(v[0],Mapping):
            for vv in v[1:]:
                assert not isinstance(vv,Mapping)
            res[k] = v[0]
        else:
            res[k] = combine_dict(*v)
    return res

class attrdict(dict):
    """A dictionary which can be accessed via attributes, for convenience"""
    def __init__(self,*a,**k):
        super(attrdict,self).__init__(*a,**k)
        self._done = set()

    def __getattr__(self,a):
        if a.startswith('_'):
            return super(attrdict,self).__getattr__(a)
        try:
            return self[a]
        except KeyError:
            raise AttributeError(a) from None
    def __setattr__(self,a,b):
        if a.startswith("_"):
            super(attrdict,self).__setattr__(a,b)
        else:
            self[a]=b
    def __delattr__(self,a):
        del self[a]

import yaml
from yaml.representer import SafeRepresenter
SafeRepresenter.add_representer(attrdict, SafeRepresenter.represent_dict)


class PathShortener:
    """This class shortens path entries so that the initial components that
    are equal to the last-used path (or the original base) are skipped.

    It is illegal to path-shorten messages whose path does not start with
    the initial prefix.

    Example: The sequence

        a b
        a b c d
        a b c e f
        a b c e g h
        a b c i 
        a b j 

    is shortened to

        0
        0 c d
        1 e f
        2 g h
        1 i
        0 j

    where the initial number is the passed-in depth.

    Usage::

        >>> d = _PathShortener(['a','b'])
        >>> d({'path': 'a b c d'.split})
        {'depth':0, 'path':['c','d']}
        >>> d({'path': 'a b c e f'.split})
        {'depth':1, 'path':['e','f']}

    etc.

    Note that the dict is modified in-place.

    """
    def __init__(self, prefix):
        self.prefix = prefix
        self.depth = len(prefix)
        self.path = []

    def __call__(self, res):
        if res.path[:self.depth] != self.prefix:
            raise RuntimeError("Wrong prefix: has %s, want %s" % (repr(res.path), repr(self.prefix)))

        p = res['path'][self.depth:]
        cdepth = min(len(p), len(self.path))
        for i in range(cdepth):
            if p[i] != self.path[i]:
                cdepth = i
                break
        self.path = p
        p = p[cdepth:]
        res['path'] = p
        res['depth'] = cdepth


class PathLongener:
    """This reverts the operation of a PathShortener.
    """
    def __init__(self, prefix):
        self.depth = len(prefix)
        self.path = prefix

    def __call__(self, res):
        p = res.pop('path', ())
        d = res.pop('depth', None)
        if d is not None:
            p = self.path[:self.depth+d] + p
        self.path = p
        res['path'] = p

class _MsgRW():
    def __init__(self, path=None, stream=None):
        if (path is None) == (stream is None):
            raise RuntimeError("You need to specify either path or stream")
        from .codec import stream_unpacker
        self.path = path
        self.stream = stream

    async def __aenter__(self):
        if self.path is not None:
            self.stream = await aopen(self.path,self._mode)
        return self

    async def __aexit__(self, *tb):
        if self.path is not None:
            await self.stream.aclose()


class MsgReader(_MsgRW):
    """Read a stream of messages from a file.
    
    Usage::

        async with MsgReader("/tmp/msgs.pack") as f:
            async for msg in f:
                process(msg)
    """

    _mode = "rb"

    def __init__(self, **kw):
        super().__init__(**kw)
        from .codec import stream_unpacker
        self.unpack = stream_unpacker()

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                msg = next(self.unpack)
            except StopIteration:
                pass
            else:
                return msg
            
            d = await self.stream.read(4096)
            if d == b"":
                raise StopAsyncIteration
            self.unpack.feed(d)


packer = None

class MsgWriter(_MsgRW):
    """Write a stream of messages from a file.
    
    Usage::

        async with MsgWriter("/tmp/msgs.pack") as f:
            for msg in some_source_of_messages():  # or "async for"
                f(msg)
    """

    _mode="wb"

    def __init__(self, buflen=65536, **kw):
        super().__init__(**kw)

        self.buf = []
        self.buflen = buflen
        self.curlen = 0
        self.excess = 0

        global packer
        if packer is None:
            from .codec import packer

    async def __aexit__(self, *tb):
        async with open_cancel_scope(shield=True):
            if self.buf:
                await self.stream.write(b''.join(self.buf))
            await super().__aexit__(*tb)

    async def __call__(self, msg):
        msg = packer(msg)
        self.buf.append(msg)
        self.curlen += len(msg)
        if self.curlen >= self.buflen-self.excess:
            buf = b''.join(self.buf)
            pos = self.buflen * int(self.curlen/self.buflen) - self.excess
            assert pos > 0
            wb,buf = buf[:pos],buf[pos:]
            self.buf = [buf]
            self.curlen = len(buf)
            self.excess = 0
            await self.stream.write(wb)

    async def flush(self):
        if self.buf:
            buf = b''.join(self.buf)
            self.buf = []
            self.excess = (self.excess+len(buf)) % self.buflen
            await self.stream.write(buf)


class TimeOnlyFormatter(logging.Formatter):
    default_time_format = "%H:%M:%S"
    default_msec_format = "%s.%03d"




import os
import logging
logger = logging.getLogger(__name__)

from yaml import safe_load

def load_cfg(cfg):
    global cfgpath
    if os.path.exists(cfg):
        pass
    elif os.path.exists(os.path.join("tests",cfg)): # pragma: no cover
        cfg = os.path.join("tests",cfg)
    elif os.path.exists(os.path.join(os.pardir,cfg)): # pragma: no cover
        cfg = os.path.join(os.pardir,cfg)
    else: # pragma: no cover
        raise RuntimeError("Config file '%s' not found" % (cfg,))

    cfgpath = cfg
    with open(cfg) as f:
        cfg = safe_load(f)

    from logging.config import dictConfig
    cfg['disable_existing_loggers'] = False
    dictConfig(cfg)
    logging.captureWarnings(True)
    logger.debug("Test %s","starting up")
    return cfg

cfg = load_cfg(os.environ.get('LOG_CFG',"logging.cfg"))





try:
    from contextlib import asynccontextmanager, AsyncExitStack
except ImportError:
    from async_generator import asynccontextmanager
    from async_exit_stack import AsyncExitStack
import anyio
import mock
import attr
import copy
import trio
import time
from pprint import pformat
from functools import partial

from distkv.client import open_client
from distkv.default import CFG
from distkv.server import Server
from distkv.codec import unpacker
from distkv.util import attrdict

import logging
logger = logging.getLogger(__name__)

otm = time.time

@asynccontextmanager
async def stdtest(n=1, run=True, client=True, tocks=20, **kw):
    TESTCFG = copy.deepcopy(CFG)
    TESTCFG.server.port = None
    TESTCFG.root="test"

    clock = trio.hazmat.current_clock()
    clock._autojump_threshold = 0.001

    @attr.s
    class S:
        tg = attr.ib()
        serfs = attr.ib(factory=set)
        splits = attr.ib(factory=set)
        s = [] # servers
        c = [] # clients

        async def ready(self, i=None):
            if i is not None:
                await self.s[i].is_ready
                return self.s[i]
            for s in self.s:
                if s is not None:
                    await s.is_ready
            return self.s

        def __iter__(self):
            return iter(self.s)

        @asynccontextmanager
        async def client(self, i:int = 0, args: dict={}):
            """Get a client for the i'th server."""
            await self.s[i].is_serving
            async with open_client(host='localhost', port=st.s[i].port, **args) as c:
                yield c

        def split(self, s):
            assert s not in self.splits
            logger.debug("Split: add %d",s)
            self.splits.add(s)

        def join(self, s):
            logger.debug("Split: join %d",s)
            self.splits.remove(s)

    async def mock_send_ping(self,old):
        assert self._tock < tocks
        await old()

    async def mock_get_host_port(st, node):
        i = int(node.name[node.name.rindex('_')+1:])
        s = st.s[i]
        await s.is_serving
        return ('localhost',s.port)

    def tm():
        try:
            return trio.current_time()
        except RuntimeError:
            return otm()

    async with anyio.create_task_group() as tg:
        st = S(tg)
        async with AsyncExitStack() as ex:
            ex.enter_context(mock.patch("time.time", new=tm))
            ex.enter_context(mock.patch("aioserf.serf_client", new=partial(mock_serf_client,st)))

            for i in range(n):
                name = "test_"+str(i)
                args = kw.get(name, kw.get('args', attrdict()))
                if 'cfg' not in args:
                    args['cfg'] = args.get('cfg',TESTCFG).copy()
                    args['cfg']['serf'] = args['cfg']['serf'].copy()
                    args['cfg']['serf']['i'] = i
                s = Server(name, **args)
                ex.enter_context(mock.patch.object(s, "_send_ping", new=partial(mock_send_ping,s,s._send_ping)))
                ex.enter_context(mock.patch.object(s, "_get_host_port", new=partial(mock_get_host_port,st)))
                st.s.append(s)
            for i in range(n):
                if kw.get("run_"+str(i), run):
                    r = anyio.create_event()
                    await tg.spawn(st.s[i].serve,r)
                    await r.wait()
            try:
                yield st
            finally:
                logger.info("Runtime: %s", clock.current_time())
                await tg.cancel_scope.cancel()
        logger.info("End")

@asynccontextmanager
async def mock_serf_client(master, **cfg):
    async with anyio.create_task_group() as tg:
        ms = MockServ(tg, master, **cfg)
        master.serfs.add(ms)
        try:
            yield ms
        finally:
            master.serfs.remove(ms)

class MockServ:
    def __init__(self, tg, master, **cfg):
        self.cfg = cfg
        self.tg = tg
        self.streams = {}
        self._master = master

    def __hash__(self):
        return id(self)

    def spawn(self, *args, **kw):
        return self.tg.spawn(*args, **kw)

    def stream(self, event_types='*'):
        if ',' in event_types or not event_types.startswith('user:'):
            raise RuntimeError("not supported")
        s = MockSerfStream(self, event_types)
        return s

    async def event(self, typ, payload, coalesce=False):
        logger.debug("SERF:%s: %s", typ, repr(unpacker(payload)))
        typ = typ[typ.index('.')+1:]
        for s in list(self._master.serfs):
            for x in self._master.splits:
                if (s.cfg.get('i',0) < x) != (self.cfg.get('i',0) < x):
                    break
            else:
                s = s.streams.get(typ, None)
                if s is not None:
                    await s.q.put(payload)

class MockSerfStream:
    def __init__(self, serf, typ):
        self.serf = serf
        self.typ = typ[typ.index('.')+1:]

    async def __aenter__(self):
        if self.typ in self.serf.streams:
            raise RuntimeError("Only one listener per event type allowed")
        self.q = anyio.create_queue(100)
        self.serf.streams[self.typ] = self
        return self

    async def __aexit__(self, *tb):
        del self.serf.streams[self.typ]
        del self.q

    def __aiter__(self):
        return self

    async def __anext__(self):
        res = await self.q.get()
        return attrdict(payload=res)




#from trio_click.testing import CliRunner
import io
import sys

from distkv.command import main

async def run(*args, expect_exit=0):
    old_stdout = sys.stdout
    sys.stdout = io.StringIO()
    try:
        res = await main.main(args)
        return res
    except SystemExit as exc:
        res = exc
        assert exc.code == expect_exit, exc.code
        return exc
    except BaseException as exc:
        res = exc
        raise
    else:
        assert expect_exit == 0
        return res
    finally:
        res.stdout = sys.stdout.getvalue()
        sys.stdout = old_stdout



"""Top-level package for DistKV."""

import pkg_resources  # part of setuptools
_version = pkg_resources.require("distkv")[0].version
del pkg_resources

_version_tuple = tuple(int(x) for x in _version.split('.'))




import pytest
import anyio
import trio
import mock
from time import time

from trio_click.testing import CliRunner
from .mock_serf import stdtest
from .run import run
from distkv.client import ServerError
import aioserf
import msgpack
from distkv.util import attrdict

import logging
logger = logging.getLogger(__name__)

N=20

@pytest.mark.trio
async def test_10_many(autojump_clock):
    """
    This test starts multiple servers at the same time.
    """
    async with stdtest(test_1={'init':420}, n=N, tocks=1000) as st:
        s = st.s[1]
        async with st.client(1) as ci:
            assert (await ci.request("get_value", path=())).value == 420
            await ci.request("set_value", path=("ping",), value="pong")

        await anyio.sleep(1)
        async def s1(i):
            async with st.client(i) as c:
                assert (await c.request("get_value", path=())).value == 420
                assert (await c.request("get_value", path=("ping",))).value == "pong"
                await c.request("set_value",path=("foo",i),value=420+i)

        async with anyio.create_task_group() as tg:
            for i in range(1,N):
                await tg.spawn(s1,i)

        await anyio.sleep(1)
        NN=min(N-1,3)
        for j in [0]+s._random.sample(range(1,N),NN):
            async with st.client(j) as c:
                for i in s._random.sample(range(1,N),NN):
                    assert (await c.request("get_value", path=("foo",i))).value == 420+i

        #await anyio.sleep(100)

@pytest.mark.trio
@pytest.mark.parametrize("tocky", [-10,-2,-1,0,1,2,10])
async def test_11_split1(autojump_clock, tocky):
    """
    This test starts multiple servers at the same time.
    """
    n_two = 0

    async with stdtest(test_1={'init':420}, n=N, tocks=1000) as st:
        async def watch():
            nonlocal n_two
            async with aioserf.serf_client() as s:
                async with s.stream("user:test.update") as sr:
                    async for r in sr:
                        msg = msgpack.unpackb(r.payload, object_pairs_hook=attrdict, raw=False, use_list=False)
                        if msg.get('value','') == 'two':
                            n_two += 1

                    i.s()

        await st.tg.spawn(watch)
        s = st.s[1]
        async with st.client(1) as ci:
            assert (await ci.request("get_value", path=())).value == 420
            await ci.request("set_value", path=("ping",), value="pong")

        async def s1(i):
            async with st.client(i) as c:
                assert (await c.request("get_value", path=())).value == 420
                await anyio.sleep(5)
                assert (await c.request("get_value", path=("ping",))).value == "pong"
                await c.request("set_value",path=("foo",i),value=420+i)

        async with anyio.create_task_group() as tg:
            for i in range(1,N):
                await tg.spawn(s1,i)

        await anyio.sleep(30)
        st.split(N//2)
        if tocky:
            async with st.client(2 if tocky < 0 else 14) as ci:
                for i in range(abs(tocky)):
                    await ci.request("set_value", path=("one",i), value="two")
        await anyio.sleep(30)

        async with st.client(N-1) as c:
            await c.request("set_value", path=("ping",), value="pongpang")
        await anyio.sleep(1)
        async with st.client(0) as c:
            assert (await c.request("get_value", path=('ping',))).value == "pong"
        st.join(N//2)
        await anyio.sleep(20)
        async with st.client(0) as c:
            assert (await c.request("get_value", path=('ping',))).value == "pongpang"

    # Now make sure that updates are transmitted once
    assert n_two <= N+1




import pytest
import anyio
import trio
import mock
from time import time

from trio_click.testing import CliRunner
from .mock_serf import stdtest
from .run import run
from distkv.client import ServerError
from distkv.util import PathLongener

import logging
logger = logging.getLogger(__name__)

@pytest.mark.trio
async def test_21_load_save(autojump_clock, tmpdir):
    """also used to check watching"""
    path = tmpdir.join("foo")
    logger.debug("START")
    msgs = []
    s = None
    async def watch_changes(c,d):
        l = PathLongener(())
        res = await c.request(action="watch", path=(), iter=True, nchain=9, state=True)
        await d.set()
        async for m in res:
            l(m)
            msgs.append(m)

    async with stdtest(args={'init':234}, tocks=30) as st:
        s, = st.s
        async with st.client() as c:
            assert (await c.request("get_value", path=())).value == 234
            d = anyio.create_event()
            await c.tg.spawn(watch_changes,c,d)
            await d.wait()

            r = await c.request("set_value", path=("foo",), value="hello", nchain=3)
            r = await c.request("set_value", path=("foo","bar"), value="baz", nchain=3)
            r = await c.request("set_value", path=(), value=2345, nchain=3)
            await anyio.sleep(1) # allow the writer to write

        logger.debug("SAVE %s",path)
        await s.save(path)
        logger.debug("SAVED")

    logger.debug("NEXT")
    for m in msgs:
        m.pop('tock',None)
    assert sorted(msgs, key=lambda x:x.chain.tick) == [
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 1},
        'path': (),
        'value': 234},
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 2},
        'path': ('foo',),
        'value': 'hello'},
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 3},
        'path': ('foo', 'bar',),
        'value': 'baz'},
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 4},
        'path': (),
        'value': 2345},
    ]

    msgs = []
    async with stdtest(run=False) as st:
        s, = st.s
        logger.debug("LOAD %s",path)
        await s.load(path, local=True)
        logger.debug("LOADED")
        r = anyio.create_event()
        await st.tg.spawn(st.s[0].serve, r)
        await r.wait()
        logger.debug("RUNNING")

        async with st.client() as c:
            d = anyio.create_event()
            await c.tg.spawn(watch_changes,c,d)
            await d.wait()

            await c.request("set_value", path=("foof",), value="again")
            assert (await c.request("get_value", path=("foo",))).value == 'hello'
            assert (await c.request("get_value", path=("foo","bar"))).value == 'baz'
            assert (await c.request("get_value", path=(()))).value == 2345
            await anyio.sleep(1) # allow the writer to write
    for m in msgs:
        m.pop('tock',None)
    assert sorted(msgs, key=lambda x:x.chain.tick) == [
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 2},
        'path': ('foo',),
        'value': 'hello'},
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 3},
        'path': ('foo', 'bar',),
        'value': 'baz'},
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 4},
        'path': (),
        'value': 2345},
        {'chain': {'node': 'test_0', 'prev': None, 'tick': 5},
        'path': ('foof',),
        'value': 'again'},
    ]
    logger.debug("OK")

@pytest.mark.trio
async def test_02_cmd(autojump_clock):
    async with stdtest(args={'init':123}) as st:
        s, = st.s
        async with st.client() as c:
            assert (await c.request("get_value", path=())).value == 123

            r = await run("client","-p",s.port,"set","-v","hello","foo")
            r = await run("client","-p",s.port,"set","-ev","'baz'","foo","bar")

            r = await run("client","-p",s.port,"get")
            assert r.stdout == "123\n"

            r = await run("client","-p",s.port,"get", "foo")
            assert r.stdout == "'hello'\n"

            r = await run("client","-p",s.port,"get", "foo", "bar")
            assert r.stdout == "'baz'\n"

            r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            assert r == {'nodes': {'test_0': 3}, 'known': {'test_0': ((1, 4),)}, 'missing': {}, 'remote_missing': {}, 'seq': 2}

            assert (await c.request("get_value", node="test_0", tick=1)).value == 123
            assert (await c.request("get_value", node="test_0", tick=2)).value == "hello"
            assert (await c.request("get_value", node="test_0", tick=3)).value == "baz"

            r = await c.request("set_value", path=(), value=1234, nchain=3)
            assert r.prev==123
            assert r.chain.tick == 4

            # does not yet exist
            with pytest.raises(ServerError):
                await c.request("get_value", node="test_0", tick=8)
            # has been superseded
            with pytest.raises(ServerError):
                await c.request("get_value", node="test_0", tick=1)
            # works
            assert (await c.request("get_value", node="test_0", tick=4)).value == 1234

            r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            assert r == {'nodes': {'test_0': 4}, 'known': {'test_0': ((1, 5),)}, 'missing': {}, 'remote_missing': {}}

@pytest.mark.trio
async def test_03_three(autojump_clock):
    async with stdtest(test_1={'init':125}, n=2, tocks=30) as st:
        s,si = st.s
        async with st.client(1) as ci:
            assert (await ci.request("get_value", path=())).value == 125

            r = await ci.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            # Various stages of integrating test_0
            assert \
                r == {'nodes': {'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                r == {'nodes': {'test_0': None, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                r == {'nodes': {'test_0': None, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {'test_0': (1,)}, 'remote_missing': {'test_0': (1,)}} or \
                r == {'nodes': {'test_1': 1, 'test_0': None,}, 'known': {'test_0': (1,), 'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                r == {'nodes': {'test_0': 0, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                False


            # This waits for test_0 to be fully up and running.
            async with st.client(0) as c:

                # At this point ci shall be fully integrated, and test_1 shall know this (mostly).
                r = await ci.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
                del r['tock']
                del r['seq']
                assert \
                        r == {'nodes': {'test_0': 0, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                        r == {'nodes': {'test_0': None, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} 

                assert (await c.request("get_value", path=())).value == 125

                r = await c.request("set_value", path=(), value=126, nchain=3)
                assert r.prev==125
                assert r.chain.tick == 1
                assert r.chain.node == 'test_0'
                assert r.chain.prev.tick == 1
                assert r.chain.prev.node == 'test_1'
                assert r.chain.prev.tick == 1

                # This verifies that the chain entry for the initial update is gone
                # and the initial change is no longer retrievable.
                # We need the latter to ensure that there are no memory leaks.
                await anyio.sleep(1)
                r = await ci.request("set_value", path=(), value=127, nchain=3)
                assert r.prev==126
                assert r.chain.tick == 2
                assert r.chain.node == 'test_1'
                assert r.chain.prev.tick == 1
                assert r.chain.prev.node == 'test_0'
                assert r.chain.prev.tick == 1
                assert r.chain.prev.prev is None

                with pytest.raises(ServerError):
                    await c.request("get_value", node="test_1", tick=1)
                with pytest.raises(ServerError):
                    await ci.request("get_value", node="test_1", tick=1)
                
                # Now test that the internal states match.
                await anyio.sleep(1)
                r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
                del r['tock']
                del r['seq']
                assert r == {'nodes': {'test_0': 1, 'test_1': 2}, 'known': {'test_0': (1,), 'test_1': ((1,3),)}, 'missing': {}, 'remote_missing': {}}

            r = await ci.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            assert r == {'nodes': {'test_0': 1, 'test_1': 2}, 'known': {'test_0': (1,), 'test_1': ((1,3),)}, 'missing': {}, 'remote_missing': {}}




# Local server
from __future__ import annotations

import anyio
from anyio.exceptions import ClosedResourceError
from anyio.abc import SocketStream, Event
from async_generator import asynccontextmanager
import msgpack
import aioserf
from typing import Any
from random import Random
import time
from range_set import RangeSet

from .model import Entry, NodeEvent, Node, Watcher, UpdateEvent
from .util import attrdict, PathShortener, PathLongener, MsgWriter, MsgReader
from . import client as distkv_client  # needs to be mock-able
from . import _version_tuple

import logging
logger = logging.getLogger(__name__)

_packer = msgpack.Packer(strict_types=False, use_bin_type=True).pack

class ClientError(ValueError):
    """report error to the client but don't dump a stack trace"""
    pass

class _NotGiven:
    pass

_client_nr = 0

def max_n(a,b):
    if a is None:
        return b
    elif b is None:
        return a
    elif a < b:
        return b
    else:
        return a

def cmp_n(a,b):
    if a is None:
        a = -1
    if b is None:
        b = -1
    return b-a

def _multi_response(fn):
    async def wrapper(self, msg):
        seq = msg.seq
        await self.send({'seq':seq, 'state':'start', 'tock':self.server.tock})
        try:
            await fn(self, msg)
        except BaseException as exc:
            raise # handled in the caller
        else:
            await self.send({'seq':seq, 'state':'end', 'tock':self.server.tock})
    return wrapper

def _stream(fn):
    """A simple wrapper that calls the original function with each message"""
    async def wrapper(self, msg):
        seq = msg.seq
        q = self.inqueue[seq]
        path = msg.get('path', None)
        longer = PathLongener(path) if path is not None else lambda x:x
        n=0
        while True:
            msg = await q.get()
            if msg.get('state','') == 'end':
                await self.send({'seq':seq, 'count':n, 'tock':self.server.tock})
                del self.inqueue[seq]
                return
            longer(msg)
            res = await fn(self, msg)
            if isinstance(res, int):
                n += res
    return wrapper

class ServerClient:
    """Represent one listening client"""
    _nursery = None

    def __init__(self, server: Server, stream: SocketStream):
        self.server = server
        self.root = server.root
        self.stream = stream
        self.seq = 0
        self.tasks = {}
        self.inqueue = {}
        self._chop_path = 0

        global _client_nr
        _client_nr += 1
        self._client_nr = _client_nr
        logger.debug("CONNECT %d %s", self._client_nr, repr(stream))
    
    async def process(self, msg, stream=False):
        """
        Process an incoming message.
        """
        seq = msg.seq
        async with anyio.open_cancel_scope() as s:
            self.tasks[seq] = s
            try:
                if 'chain' in msg:
                    msg.chain = NodeEvent.deserialize(msg.chain)

                try:
                    fn = getattr(self, ('scmd_' if stream else 'cmd_') + str(msg.action))
                except AttributeError:
                    raise ClientError("Command not recognized: " + repr(msg.action))
                else:
                    res = await fn(msg)
                    if res is not None: await self.send_result(seq, res)

            except BrokenPipeError as exc:
                logger.error("ERR%d: %s", self._client_nr, repr(exc))

            except Exception as exc:
                if not isinstance(exc, (KeyError,ClientError)):
                    logger.exception("ERR%d: %s", self._client_nr, repr(msg))
                await self.send({'error': str(exc), 'seq': seq, 'tock':self.server.tock})

            finally:
                del self.tasks[seq]
        
    async def cmd_root(self, msg):
        """Change to a sub-tree.
        """
        entry = self.root.follow(*msg.path)
        self.root = entry
        self._chop_path += len(msg.path)
        return entry.serialize(chop_path=self._chop_path)

    async def cmd_get_value(self, msg):
        """Get a node's value.
        """
        if 'node' in msg and 'path' not in msg:
            n = Node(msg.node, cache=self.server._nodes, create=False)
            return n[msg.tick].serialize(chop_path=self._chop_path, nchain=msg.get('nchain', 0))

        try:
            entry = self.root.follow(*msg.path, create=False)
        except KeyError:
            entry = {'value': None}
        else:
            entry = entry.serialize(chop_path=-1, nchain=msg.get('nchain', 0))
        return entry

    async def cmd_set_value(self, msg):
        """Set a node's value.
        """
        entry = self.root.follow(*msg.path)
        send_prev = True
        if 'prev' in msg:
            if entry.data != msg.prev:
                raise ClientError("Data is %s" % (repr(entry.data),))
            send_prev = False
        if 'chain' in msg:
            if entry.chain is not None:
                if msg.chain is None:
                    raise ClientError("This entry already exists")
                if entry.chain != msg.chain:
                    raise ClientError("Chain is %s" % (repr(entry.chain),))
            elif msg.chain is not None:
                raise ClientError("This entry is new")
            send_prev = False
        res = {'changed': entry.data != msg.value}
        if send_prev:
            res['prev'] = entry.data

        async with self.server.next_event() as event:
            await entry.set_data(event, msg.value, dropped=self.server._dropper)

        nchain = msg.get('nchain', 1)
        if nchain > 0:
            res['chain'] = entry.chain.serialize(nchain=nchain)
        return res

    @_stream
    async def scmd_update(self, msg):
        """
        Apply a stored update.

        This streams the input.
        """
        msg = UpdateEvent.deserialize(self.root, msg)
        await msg.entry.apply(msg, dropped=self._dropper)

    async def cmd_update(self, msg):
        """
        Apply a stored update.

        You usually do this via a stream command.
        """
        msg = UpdateEvent.deserialize(self.root, msg)
        res = await msg.entry.apply(msg, dropped=self._dropper)
        if res is None:
            return False
        else:
            return res.serialize(chop_path=self._chop_path)

    async def cmd_delete_value(self, msg):
        """Delete a node's value.
        Sub-nodes are not affected.
        """
        if 'value' in msg:
            raise ClientError("A deleted entry can't have a value")
        msg.value = None
        return await self.cmd_set_value(msg)

    @_multi_response
    async def cmd_get_tree(self, msg):
        try:
            entry = self.root.follow(*msg.path, create=False)
        except KeyError:
            await self.send_result(seq, {'value':None})
            return

        seq = msg.seq
        nchain = msg.get('nchain',0)
        shorter = PathShortener(entry.path)

        async def sender(entry):
            if entry.data is None:
                return
            res = entry.serialize(chop_path=self._chop_path, nchain=nchain)
            await self.send_result(seq, res)
        await entry.walk(sender)

    def cmd_get_state(self, msg):
        """Return some info about this node's internal state"""
        return self.server.get_state(**msg)

    @_multi_response
    async def cmd_watch(self, msg):
        """Monitor a subtree for changes.
        If ``state`` is set, dump the initial state before reporting them.
        """
        entry = self.root.follow(*msg.path, create=True)
        seq = msg.seq
        nchain = msg.get('nchain',None)
        if nchain is None:
            nchain = self.cfg['change']['length'] if create else 0

        async with Watcher(entry) as watcher:
            shorter = PathShortener(entry.path)
            if msg.get('state', False):
                async def sender(entry):
                    res = entry.serialize(nchain=nchain)
                    shorter(res)
                    await self.send_result(seq, res)
                await entry.walk(sender)

            async for msg in watcher:
                res = msg.entry.serialize(chop_path=self._chop_path, nchain=nchain)
                shorter(res)
                await self.send_result(seq, res)


    async def cmd_delete_tree(self, msg):
        """Delete a node's value.
        Sub-nodes are cleared (after their parent).
        """
        seq = msg.seq
        nchain = msg.get('nchain', 0)
        if nchain:
            await self.send({'seq':seq, 'state':'start', 'tock':self.server.tock})
        ps = PathShortener(msg.path)

        try:
            entry = self.root.follow(*msg.path)
        except KeyError:
            return False

        async def _del(entry):
            res = 0
            if entry.data is not None:
                async with self.server.next_event() as event:
                    evt = await entry.set_data(event, None, dropped=self.server._dropper)
                    if nchain:
                        r = evt.serialize(chop_path=self._chop_path, nchain=nchain, with_old=True)
                        r['seq'] = seq
                        del r['new_value']  # always None
                        ps(r)
                        await self.send(r)
                res += 1
            for v in entry.values():
                res += await _del(v)
            return res
        res = await _del(entry)
        if nchain:
            await self.send({'seq':seq, 'state':'end', 'tock':self.server.tock})
        else:
            return {'changed': res}

    async def cmd_log(self, msg):
        await self.server.run_saver(path=msg.path, save_state=msg.get('state', False))
        return True

    async def cmd_save(self, msg):
        await self.server.save(path=msg.path)
        return True

    async def cmd_stop(self, msg):
        try:
            t = self.tasks[msg.task]
        except KeyError:
            return False
        t.cancel()
        return True

    def send(self, msg):
        logger.debug("OUT%d: %s", self._client_nr, msg)
        return self.stream.send_all(_packer(msg))

    def send_result(self, seq, res):
        res['seq'] = seq
        res['tock'] = self.server.tock
        return self.send(res)

    async def run(self):
        """Main loop for this client connection."""
        unpacker = msgpack.Unpacker(object_pairs_hook=attrdict, raw=False, use_list=False)

        async with anyio.create_task_group() as tg:
            self.tg = tg
            await self.send({'seq': 0, 'version': _version_tuple,
                'node':self.server.node.name, 'tick':self.server.node.tick,
                'tock':self.server.tock})

            while True:
                for msg in unpacker:
                    try:
                        logger.debug("IN %d: %s", self._client_nr, msg)
                        seq = msg.seq
                        q = self.inqueue.get(seq, None)
                        if q is not None:
                            await q.put(msg)
                        elif msg.get('state','') == 'start':
                            q = anyio.create_queue(100)
                            self.inqueue[seq] = q
                            await self.tg.spawn(self.process, q, True)
                        else:
                            if self.seq >= seq:
                                raise ClientError("Sequence numbers are not monotonic: %d < %d" % (self.seq, msg.seq))
                            self.seq = seq
                            await self.tg.spawn(self.process, msg)
                    except Exception as exc:
                        if not isinstance(exc, ClientError):
                            logger.exception("ERR %d: Client error on %s", self._client_nr, repr(msg))
                        await self.send({'error': str(exc), 'tock':self.server.tock})

                try:
                    buf = await self.stream.receive_some(4096)
                except (ConnectionResetError, ClosedResourceError):
                    return  # closed/reset/whatever
                if len(buf) == 0:  # Connection was closed.
                    return # done
                unpacker.feed(buf)


class Server:
    serf = None
    _ready = None

    def __init__(self, name: str, cfg: dict, init: Any = _NotGiven, root: Entry = None):
        if root is None:
            root = Entry("ROOT", None)
        self.root = root
        self.cfg = cfg
        self._nodes = {}
        self.node = Node(name, None, cache=self._nodes)
        self._tock = 0
        self._init = init

        self._evt_lock = anyio.create_lock()
        self._random = Random()

    @asynccontextmanager
    async def next_event(self):
        """A context manager which returns the next event under a lock.

        This increments ``tock`` because that increases the chance that the
        node (or split) where something actually happens wins a collision.
        """
        async with self._evt_lock:
            self.node.tick += 1
            self._tock += 1
            yield NodeEvent(self.node)
            self._tock += 1

    @property
    def tock(self):
        """Retrieve ``tock``.
        
        Also increments it because tock values may not be re-used."""
        self._tock += 1
        return self._tock

    @property
    def random(self):
        """An attribute that generates a random fraction in ``[0,1)``."""
        return self._random.random()

    def _dropper(self, evt, old_evt=_NotGiven):
        """Drop either one event, or any event that is in ``old_evt`` but not in
        ``evt``."""
        if old_evt is None:
            return
        if old_evt is _NotGiven:
            evt.node.supersede(evt.tick)
            return

        nt = {}
        while evt is not None:
            nt[evt.node.name] = evt.tick
            evt = evt.prev
        while old_evt is not None:
            if nt.get(old_evt.node.name, 0) != old_evt.tick:
                old_evt.node.supersede(old_evt.tick)
            old_evt = old_evt.prev


    async def _send_event(self, action, msg, coalesce=False):
        msg['tock'] = self.tock
        if 'node' not in msg:
            msg['node'] = self.node.name
        if 'tick' not in msg:
            msg['tick'] = self.node.tick
        msg = _packer(msg)
        await self.serf.event(self.cfg['root']+'.'+action, msg, coalesce=coalesce)

    async def watcher(self):
        """This method implements the task that watches a (sub)tree for changes"""
        async with Watcher(self.root) as watcher:
            async for msg in watcher:
                if msg.event.node != self.node:
                    continue
                if self.node.tick is None:
                    continue
                p = msg.serialize(nchain=self.cfg['change']['length'])
                await self._send_event('update', p)

    async def _process(self, msg):
        """Dispatch an incoming message to ``cmd_*`` methods"""
        try:
            fn = getattr(self, 'cmd_' + str(msg.action))
        except AttributeError:
            raise ClientError("Command not recognized: " + repr(msg.action))
        else:
            return await fn(msg)

    async def get_state(self, nodes=False, known=False, missing=False,
            remote_missing=False, **kw):
        """Return some info about this node's internal state"""
        res = attrdict()
        if nodes:
            nd = res.nodes = {}
            for n in self._nodes.values():
                nd[n.name] = n.tick
        if known:
            nd = res.known = {}
            for n in self._nodes.values():
                lk = n.local_known
                if len(lk):
                    nd[n.name] = lk.__getstate__()
        if missing:
            nd = res.missing = {}
            for n in self._nodes.values():
                if not n.tick:
                    continue
                lk = n.local_missing
                if len(lk):
                    nd[n.name] = lk.__getstate__()
        if remote_missing:
            nd = res.remote_missing = {}
            for n in self._nodes.values():
                lk = n.remote_missing
                if len(lk):
                    nd[n.name] = lk.__getstate__()
        return res

    async def user_update(self, msg):
        """Process an update."""
        msg = UpdateEvent.deserialize(self.root, msg, cache=self._nodes)
        await msg.entry.apply(msg, dropped=self._dropper)

    async def user_info(self, msg):
        """Process info broadcasts.
        
        These are mainly used in the split recovery protocol."""

        if msg.node == self.node.name:
            return  # ignore our own message

        # 'reason' is the ping chain from the node that triggered this info
        # message. If we're on "our" chain but not on the "other", then we
        # also need to start recovery. Also, if the "other" chain is better,
        # we need to replace ours.
        # This happens when, immediately after a split, our ping is
        # rejected by the remote side. Instead of sending a ping of their
        # own immediately, they send their chain as reason with
        # replace=True.
        r = msg.get('reason', None)
        if r is not None and self._recover_task is None:
            replace = r.get('replace', False)
            pos = self.last_ping_evt.find(self.node)
            rev = NodeEvent.deserialize(r, cache=self._nodes, check_dup=False)
            if replace:
                r.tock = msg.tock
                self.last_ping = r
                self.last_ping_evt = rev
            if pos is not None:
                self.sane_ping = self.last_ping_evt
                await self.spawn(self.recover_split, pos)

        # Step 1
        ticks = msg.get('ticks', None)
        if ticks is not None:
            for n,t in ticks.items():
                n = Node(n, cache=self._nodes)
                n.tick = max_n(n.tick, t)
            if self._recover_event1 is not None and \
                    (self.sane_ping is None or self.node in self.sane_ping):
                logger.debug("Step1 %s: triggered by %s", self.node.name, self.sane_ping.serialize() if self.sane_ping else "-")
                await self._recover_event1.set()
            elif self._recover_event1 is not None:
                logger.debug("Step1 %s: not in %s", self.node.name, self.sane_ping.serialize())
#           else:
#               logger.debug("Step1 %s: no event", self.node.name)

        # Step 2
        missing = msg.get('missing', None)
        if missing is not None:
            nn = 0
            for n,k in missing.items():
                n = Node(n, cache=self._nodes)
                r = RangeSet()
                r.__setstate__(k)
                nn += len(r)
                n.reported_missing(r)
                mr = self.seen_missing.get(n, None)
                if mr is None:
                    self.seen_missing[n] = r
                else:
                    mr += r
            if self._recover_event2 is not None and \
                    (self.sane_ping is None or self.node in self.sane_ping):
                logger.debug("Step2 %s: triggered by %s", self.node.name, self.sane_ping.serialize() if self.sane_ping else "-")
                await self._recover_event2.set()
            elif self._recover_event2 is not None:
                logger.debug("Step2 %s: not in %s", self.node.name, self.sane_ping.serialize())
#           else:
#               logger.debug("Step2 %s: no event", self.node.name)

            if nn > 0:
                await self._run_send_missing()

        # Step 3
        known = msg.get('known',None)
        if known is not None:
            for n,k in known.items():
                n = Node(n, cache=self._nodes)
                r = RangeSet()
                r.__setstate__(k)
                n.reported_known(r)



    async def user_ping(self, msg):
        """Process ping broadcasts.

        Just queue them for the ``pinger`` task to handle.
        """
        await self.ping_q.put(msg)

    async def monitor(self, action: str, delay: Event = None):
        """The task that hooks to Serf's event stream for receiving messages.
        
        Args:
          ``action``: The action name, corresponding to a ``user_*`` method.
          ``delay``: an optional event to wait for, between starting the
                     listener and actually processing messages. This helps
                     to avoid possible inconsistency errors on startup.
        """
        cmd = getattr(self, 'user_'+action)
        async with self.serf.stream('user:%s.%s' % (self.cfg['root'], action)) as stream:
            if delay is not None:
                await delay.wait()

            async for resp in stream:
                msg = msgpack.unpackb(resp.payload, object_pairs_hook=attrdict, raw=False, use_list=False)
                self.tock_seen(msg.get('tock',0))
                await cmd(msg)

    def tock_seen(self, tock):
        """Update the current ``tock`` if it's not high enough."""
        self._tock = max(self._tock, tock)

    async def _send_ping(self):
        """Send a ping message and note when to send the next one,
        assuming that no intervening ping arrives.
        """
        if self._tock > 1000:
            raise RuntimeError("This test has gone on too long")

        self.last_ping_evt = msg = NodeEvent(self.node, prev=self.last_ping_evt)
        self.last_ping = msg = msg.serialize(self.cfg['ping']['length'])
        await self._send_event('ping', msg)

        t = time.time()
        self.next_ping = t + self._time_to_next_ping() * self.cfg['ping']['clock']

    def _time_to_next_ping(self):
        """Calculates the time until sending the next ping is a good idea,
        assuming that none arrive in the meantime, in clocks."""
        if self.last_ping_evt.node == self.node:
            # we transmitted the last ping. If no other ping arrives we are
            # the only node.
            return 3
        # check whether the first half of the ping chain contains nonzero ticks
        # so that if we're not fully up yet, the chain doesn't only consist of
        # nodes that don't work.
        c = self.last_ping_evt.prev
        p = s = 0
        l = 1
        while c is not None:
            if c.tick is not None and c.tick > 0 and p == 0:
                p = l
            if c.node == self.node:
                s = l
            l += 1
            c = c.prev
        if not self._ready.is_set():
            if p > l//2:
                # No it does not. Do not participate.
                return 3

        if s > 0:
            # We are on the chain. Send ping depending on our position.
            return 2 - (s-1)/l
            # this will never be 1 because we need to leave some time for
            # interlopers, below. Otherwise we could divide by l-1, as
            # l must be at least 2. s must also be at least 1.

        if l < self.cfg['ping']['length']-1:
            # the chain is too short. Try harder to get onto it.
            f = 3
        else:
            f = 10
        if self.random < 1/f/len(self._nodes):
            # send early (try getting onto the chain)
            return 1+self.random/len(self._nodes)
        else:
            # send late (fallback)
            return 2.5+self.random/2

    async def pinger(self, delay):
        """
        This task
        * sends PING messages
        * handles incoming pings
        * triggers split recovery

        The initial ping is delayed randomly.

        Args:
          ``delay``: an event to set after the initial ping message has
                     been sent.
        """
        clock = self.cfg['ping']['clock']

        # initial delay: anywhere from clock/2 to clock seconds
        await anyio.sleep((self.random/2+0.5)*clock)
        await self._send_ping()
        await delay.set()

        while True:
            msg = None
            t = max(self.next_ping - time.time(), 0)
            #logger.debug("S %s: wait %s", self.node.name, t)
            async with anyio.move_on_after(t):
                msg = await self.ping_q.get()
            if msg is None:
                await self._send_ping()
                continue

            # Handle incoming ping
            event = NodeEvent.deserialize(msg, cache=self._nodes, check_dup=False)

            if self.node == event.node:
                # my own message, returned
                continue

            if event.prev is not None and self.last_ping_evt.equals(event.prev):
                # valid "next" ping
                self.last_ping = msg
                self.last_ping_evt = event
                self.next_ping = time.time() + clock * self._time_to_next_ping()
                continue

            saved_ping = self.last_ping_evt
            # colliding pings.
            #
            # This while loop is only used as a "goto forward".
            # ``break`` == "the new ping is better"
            # ``pass``  == "the last ping I saw is better"
            while True:
                if event.tick is None and self._ready.is_set():
                    # always prefer our ping
                    break
                if event.tick is not None and not self._ready.is_set():
                    # always prefer the other ping
                    pass
                else:
                    if msg.tock < self.last_ping.tock:
                        break
                    if msg.tock == self.last_ping.tock:
                        if cmp_n(event.tick, self.last_ping_evt.tick) < 0:
                            break
                        if cmp_n(event.tick, self.last_ping_evt.tick) == 0:
                            if event.node.name < self.last_ping_evt.node.name:
                                break
                            assert event.node.name != self.last_ping_evt.node.name

                # If we get here, the other ping is "better".
                logger.debug("Coll Ack %s: %s",self.node.name,msg)
                self.last_ping = msg
                self.last_ping_evt = event
                t = time.time()
                self.next_ping = time.time() + clock * self._time_to_next_ping()
                break  # always terminate the loop

            if event.prev is not None and event.prev.equals(self.last_ping_evt):
                # These pings refer to the same previous ping. Good.
                logger.debug("Coll PRE %s: %s",self.node.name,msg)
                continue

            if self.last_ping is not msg:
                logger.debug("Coll NO  %s: %s",self.node.name,msg)


            # We either have a healed network split (bad) or are new (oh well).
            if self._ready.is_set():
                # otherwise I have nothing to say

                if event.tick is None:
                    # The colliding node does not have data. Ignore.
                    continue

                if saved_ping.tick is None:
                    # The node sending the last ping had no data, so
                    # there's no collision now.
                    continue

                pos = saved_ping.find(self.node)
                if pos is not None:
                    if self._recover_task is None:
                        # await self._recover_task.cancel()
                        if self.sane_ping is None:
                            self.sane_ping = saved_ping
                        await self.spawn(self.recover_split, pos, self.last_ping is not msg)
            elif self.fetch_running is None and self.last_ping_evt.tick is not None:
                await self.spawn(self.fetch_data)

    async def _get_host_port(self, node):
        """Retrieve the remote system to connect to"""
        port = self.cfg['server']['port']
        domain = self.cfg['domain']
        host = node.name
        if domain is not None:
            host += '.'+domain
        return (host,port)

    async def do_send_missing(self):
        """Task to periodically send "missing …" messages
        """
        logger.debug("send-missing %s started", self.node.name)
        clock = self.cfg['ping']['clock']/2
        while self.fetch_missing:
            if self.fetch_running is not False:
                logger.debug("send-missing %s halted", self.node.name)
                return
            clock *= self.random/2+1
            await anyio.sleep(clock)

            n = 0
            msg = dict()
            for n in list(self.fetch_missing):
                m = n.local_missing
                nl = len(m)
                if nl == 0:
                    self.fetch_missing.remove(n)
                    continue

                mr = self.seen_missing.get(n.name, None)
                if mr is not None:
                    m -= mr
                if len(m) == 0:
                    continue
                msg[n.name] = m.__getstate__()
            self.seen_missing = {}
            if not n:  # nothing more to do
                break
            if not len(msg):  # others already did the work, this time
                continue
            msg = attrdict(missing=msg)
            await self._send_event('info', msg)

        logger.debug("send-missing %s ended", self.node.name)
        if self.node.tick is None:
            self.node.tick = 0
            await self._check_ticked()
        self.fetch_running = None

    async def fetch_data(self):
        """
        We are newly started and don't have any data.

        Try to get the initial data from some other node.
        """
        if self.fetch_running is not None:
            return
        self.fetch_running = True
        while True:
            n = self.last_ping_evt
            while n is not None:
                node = n.node
                n = n.prev
                if node.tick is None:  # not ready
                    continue

                try:
                    host, port = await self._get_host_port(node)
                    async with distkv_client.open_client(host, port) as client:

                        res = await client.request('get_tree', iter=True, from_server=self.node.name, nchain=999, path=())
                        async for r in res:
                            r = UpdateEvent.deserialize(self.root, r, cache=self._nodes)
                            await r.entry.apply(r, dropped=self._dropper)
                        self.tock_seen(res.end_msg.tock)

                        res = await client.request('get_state', nodes=True, from_server=self.node.name, known=True, iter=False)
                        await self._process_info(res)

                except (AttributeError, KeyError, ValueError, AssertionError, TypeError):
                    raise
                except Exception as exc:
                    logger.exception("Unable to connect to %s" % (node,))
                else:
                    # At this point we successfully cloned some other
                    # node's state, so we now need to find whatever that
                    # node didn't have.

                    for n in self._nodes.values():
                        if n.tick and len(n.local_missing):
                            self.fetch_missing.add(n)
                    if len(self.fetch_missing):
                        self.fetch_running = False
                        await self.spawn(self.do_send_missing)
                    else:
                        self.node.tick = 0
                        self.fetch_running = None
                        await self._check_ticked()
                    return
            await anyio.sleep(self.cfg['ping']['clock']*1.1)

    async def _process_info(self, msg):
        for nn,t in msg.get('nodes',{}).items():
            nn = Node(nn, cache=self._nodes)
            nn.tick = max_n(nn.tick,t)
        for nn,k in msg.get('known',{}).items():
            nn = Node(nn, cache=self._nodes)
            r = RangeSet()
            r.__setstate__(k)
            nn.reported_known(r, local=True)
        for nn,k in msg.get('remote_missing',{}).items():
            # used when loading data from a state file
            nn = Node(nn, cache=self._nodes)
            r = RangeSet()
            r.__setstate__(k)
            nn.reported_missing(r)

    async def _check_ticked(self):
        if self._ready is None:
            return
        if self.node.tick is not None:
            logger.debug("Ready %s",self.node.name)
            await self._ready.set()

    async def recover_split(self, pos, replace=False):
        """
        Recover from a network split.
        """
        clock = self.cfg['ping']['clock']
        tock = self.tock
        self._recover_tock = tock
        self._recover_event1 = anyio.create_event()
        self._recover_event2 = anyio.create_event()
        logger.info("SplitRecover %s: %s @%d", self.node.name, pos, tock)

        async with anyio.open_cancel_scope() as s:
            self._recover_task = s

            try:
                # Step 1: send an info/ticks message
                # for pos=0 this fires immediately. That's intentional.
                async with anyio.move_on_after(clock * (1-1/(1<<pos))/2) as x:
                    await self._recover_event1.wait()
                if self.sane_ping is None:
                    logger.info("SplitRecover %s: no sane 1", self.node.name)
                    return
                if x.cancel_called:
                    logger.info("SplitRecover %s: no signal 1", self.node.name)
                    msg = dict((x.name,x.tick) for x in self._nodes.values())

                    msg = attrdict(ticks=msg)
                    msg.reason = self.sane_ping.serialize()
                    msg.reason.replace = replace

                    await self._send_event('info', msg)

                # Step 2: send an info/missing message
                # for pos=0 this fires after clock/2, so that we get a
                # chance to wait for other info/ticks messages. We can't
                # trigger on them because there may be more than one, for a
                # n-way merge.
                async with anyio.move_on_after(clock * (2-1/(1<<pos))/2) as x:
                    await self._recover_event2.wait()

                if x.cancel_called:
                    logger.info("SplitRecover %s: no signal 2", self.node.name)
                    msg = dict()
                    for n in self._nodes.values():
                        if not n.tick:
                            continue
                        m = n.local_missing
                        mr = self.seen_missing.get(n.name, None)
                        if mr is not None:
                            m -= mr
                        if len(m) == 0:
                            continue
                        msg[n.name] = m.__getstate__()
                        if mr is None:
                            self.seen_missing[n.name] = m
                        else:
                            mr += m

                    msg = attrdict(missing=msg)
                    await self._send_event('info', msg)

                # wait a bit more before continuing. Again this depends on
                # `pos` so that there won't be two nodes that send the same
                # data at the same time, hopefully.
                await anyio.sleep(clock * (1-1/(1<<pos)))

                # Step 3: start a task that sends stuff
                await self._run_send_missing()

            finally:
                # Protect against cleaning up when another recovery task has
                # been started (because we saw another merge)
                if self._recover_tock != tock:
                    logger.info("SplitRecover %s: canceled @%d", self.node.name, tock)
                    return
                logger.info("SplitRecover %s: finished @%d", self.node.name, tock)
                self._recover_tock = 0
                self._recover_task = None
                self._recover_event1 = None
                self._recover_event2 = None
                self.sane_ping = None
                self.seen_missing = {}

    async def _run_send_missing(self):
        """Start :meth:`_send_missing_data` if it's not running"""

        pos = (self.sane_ping or self.last_ping_evt).find(self.node)
        if self.sending_missing is None:
            self.sending_missing = True
            await self.spawn(self._send_missing_data, pos)
        elif not self.sending_missing:
            self.sending_missing = True

    async def _send_missing_data(self, pos):
        """Step 3 of the re-join protocol.
        For each node, collect events that somebody has reported as missing,
        and re-broadcast them. If the event is unavailable, send a "known"
        message.
        """
        clock = self.cfg['ping']['clock']
        if pos is None:
            await anyio.sleep(clock*(1/2+self.random/5))
        else:
            await anyio.sleep(clock*(1-1/(1<<pos))/2)

        while self.sending_missing:
            self.sending_missing = False
            nodes = list(self._nodes.values())
            self._random.shuffle(nodes)
            known = {}
            for n in nodes:
                k = RangeSet()
                for r in n.remote_missing & n.local_known:
                    for t in range(*r):
                        if t not in n.remote_missing:
                            # some other node could have sent this while we worked
                            await anyio.sleep(self.cfg['ping']['clock']/10)
                            continue
                        if t in n:
                            msg = n[t].serialize()
                            await self._send_event('update', msg)
                        else:
                            k.add(t)
                if k:
                    known[n.name] = k.__getstate__()
                rm = n.remote_missing
                rm -= n.local_known
                assert rm is n.remote_missing
            if known:
                await self._send_event('info', attrdict(known=known))
        self.sending_missing = None

    async def load(self, path:str, local: bool = False):
        """Load data from this stream
        
        Args:
          ``fd``: The stream to read.
          ``local``: Flag whether this file contains initial data and thus
                     its contents shall not be broadcast. Don't set this if
                     the server is already operational.
        """
        longer = PathLongener(())

        if local and self.node.tick is not None:
            raise RuntimeError("This server already has data.")
        elif not local and self.node.tick is None:
            raise RuntimeError("This server is not yet operational.")
        async with MsgReader(path=path) as rdr:
            async for m in rdr:
                if 'value' in m:
                    longer(m)
                    m = UpdateEvent.deserialize(self.root, m, cache=self._nodes)
                    await m.entry.apply(m, local=local, dropped=self._dropper)
                elif 'nodes' in m or 'known' in m:
                    await self._process_info(m)
                else:
                    logger.warn("Unknown message in stream: %s", repr(m))
        logger.info("Loading finished.")
        
    async def _save(self, writer, shorter, nchain=99):
        """Save the current state.

        TODO: Add code for snapshotting.
        """
        async def saver(entry):
            res = entry.serialize(nchain=nchain)
            shorter(res)
            await writer(res)
        msg = await self.get_state(nodes=True, known=True)
        await writer(msg)
        await self.root.walk(saver)

    async def save(self, path:str=None, stream=None, delay:Event = None):
        """Save the current state to ``path`` or ``stream``."""
        shorter = PathShortener([])
        async with MsgWriter(path=path, stream=stream) as mw:
            await self._save(mw, shorter)

    async def save_stream(self, path:str=None, stream=None, save_state:bool=False, done:int=None):
        """Save the current state to ``path`` or ``stream``.
        Continue writing updates until cancelled.
        """
        shorter = PathShortener([])

        async with MsgWriter(path=path, stream=stream) as mw:
            async with Watcher(self.root) as updates:
                await self._ready.wait()

                if save_state:
                    await self._save(mw, shorter)

                msg = await self.get_state(nodes=True, known=True)
                await mw(msg)
                await mw.flush()
                if done is not None:
                    await done.set()

                async for msg in updates:
                    await mw(msg.serialize())
    
    _saver_prev = None
    async def _saver(self, path:str, done, save_state=False):

        async with anyio.open_cancel_scope() as s:
            self._saver_prev = s
            try:
                await self.save_stream(path=path, done=done, save_state=save_state)
            finally:
                if self._saver_prev is s:
                    self._saver_prev = None

    async def run_saver(self, path: str=None, stream=None, save_state=False):
        """Start a task that continually saves to disk.

        Only one saver can run at a time; if a new one is started, 
        the old one is stopped as soon as the new saver's current state is on disk.
        """
        done = anyio.create_event()
        s = self._saver_prev
        await self.spawn(self._saver, path=path, stream=stream, done=done)
        await done.wait()
        if s is not None:
            await s.cancel()

    @property
    async def is_ready(self):
        """Await this to determine if/when the server is operational."""
        await self._ready.wait()

    @property
    async def is_serving(self):
        """Await this to determine if/when the server is serving clients."""
        await self._ready2.wait()

    async def serve(self, setup_done: Event = None, log_stream = None):
        """Task that opens a Serf connection and actually runs the server.
        
        Args:
          ``setup_done``: optional event that's set when the server is initially set up.
          ``log_stream``: a binary stream to write changes and initial state to.
        """
        async with aioserf.serf_client(**self.cfg['serf']) as serf:
            # Collect all "info/missing" messages seen since the last
            # healed network split so that they're only sent once.
            self.seen_missing = {}

            # Is the missing-items-sender running?
            # None=no, otherwise flag whether it should run another round
            self.sending_missing = None

            # Nodes which list missing events
            self.fetch_missing = set()

            # Flag whether do_fetch_missing is running (True)
            # or do_send_missing is running (False)
            # or neither (None)
            self.fetch_running = None

            # Set when self.node.tick is no longer None, i.e. we have some
            # reasonable state
            self._ready = anyio.create_event()

            # set when we're ready to accept client connections
            self._ready2 = anyio.create_event()

            self.serf = serf
            self.spawn = serf.spawn

            # Queue for processing incoming ping broadcasts
            self.ping_q = anyio.create_queue(self.cfg['ping'].length+2)

            # Last "reasonable" ping seen
            self.last_ping = None
            self.last_ping_evt = None
            self.last_sent_ping = None

            # Last ping on "our" branch when a network split is healed
            self.sane_ping = None

            # Sync recovery steps so that only one node per branch answers
            self._recover_event1 = None
            self._recover_event2 = None

            # Cancel scope; if :meth:`recover_split` is running, use that
            # to cancel
            self._recover_task = None
            self._recover_tock = 0

            # used to sync starting up everything so no messages get either
            # lost, or processed prematurely
            delay = anyio.create_event()
            delay2 = anyio.create_event()

            if setup_done is not None:
                await setup_done.set()

            if self.cfg['state'] is not None:
                await self.spawn(self.save, self.cfg['state'])

            if log_stream is not None:
                await self.run_saver(stream=log_stream, save_state=True)

            # Link up our "user_*" code
            for d in dir(self):
                if d.startswith("user_"):
                    await self.spawn(self.monitor, d[5:], delay)

            await self.spawn(self.watcher)

            if self._init is not _NotGiven:
                assert self.node.tick is None
                self.node.tick = 0
                async with self.next_event() as event:
                    await self.root.set_data(event, self._init)
            
            # send initial ping
            await self.spawn(self.pinger, delay2)
            await delay2.wait()

            await anyio.sleep(0.1)
            await delay.set()
            await self._check_ticked()  # when _init is set

            cfg_s = self.cfg['server'].copy()
            cfg_s.pop('host', 'localhost')

            await self._ready.wait()
            if cfg_s.get('port',_NotGiven) is None:
                del cfg_s['port']
            async with await anyio.create_tcp_server(**cfg_s) as server:
                self.port = server.port
                logger.debug("S %s: opened port %s", self.node.name, self.port)
                await self._ready2.set()
                async for client in server.accept_connections():
                    await self.spawn(self._connect, client)

    async def _connect(self, stream):
        c = ServerClient(server=self, stream=stream)
        try:
            await c.run()
        finally:
            await stream.close()




import pytest
import anyio
import trio
import mock
from time import time

from trio_click.testing import CliRunner
from .mock_serf import stdtest
from .run import run
from distkv.client import ServerError

import logging
logger = logging.getLogger(__name__)

# This is a basic Trio test which we keep around to check that
# (a) the autojump clock works as advertised
# (b) we can use anyio.
@pytest.mark.trio
async def test_00_anyio_clock(autojump_clock):
    assert trio.current_time() == 0
    t = time()

    for i in range(10):
        start_time = trio.current_time()
        await anyio.sleep(i)
        end_time = trio.current_time()

        assert end_time - start_time == i
    assert time()-t < 1

@pytest.mark.trio
async def test_00_runner(autojump_clock):
    with pytest.raises(AssertionError):
        await run('--doesnotexist')
    await run('--doesnotexist', expect_exit=2)
    #await run('pdb','pdb')  # used for verifying that debugging works

@pytest.mark.trio
async def test_01_basic(autojump_clock):
    async with stdtest(args={'init':123}) as st:
        s, = st.s
        async with st.client() as c:
            assert (await c.request("get_value", path=())).value == 123

            r = await c.request("set_value", path=("foo"), value="hello", nchain=3)
            r = await c.request("set_value", path=("foo","bar"), value="baz", nchain=3)
            r = await c.request("get_value", path=())
            assert r.value == 123

            r = await c.request("get_value", path=("foo"))
            assert r.value == "hello"

            r = await c.request("get_value", path=("foo","bar"))
            assert r.value == "baz"

            r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            assert r == {'nodes': {'test_0': 3}, 'known': {'test_0': ((1, 4),)}, 'missing': {}, 'remote_missing': {}, 'seq': 7}

            assert (await c.request("get_value", node="test_0", tick=1)).value == 123
            assert (await c.request("get_value", node="test_0", tick=2)).value == "hello"
            assert (await c.request("get_value", node="test_0", tick=3)).value == "baz"

            r = await c.request("set_value", path=(), value=1234, nchain=3)
            assert r.prev==123
            assert r.chain.tick == 4

            # does not yet exist
            with pytest.raises(ServerError):
                await c.request("get_value", node="test_0", tick=8)
            # has been superseded
            with pytest.raises(ServerError):
                await c.request("get_value", node="test_0", tick=1)
            # works
            assert (await c.request("get_value", node="test_0", tick=4)).value == 1234

            r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            assert r == {'nodes': {'test_0': 4}, 'known': {'test_0': ((1, 5),)}, 'missing': {}, 'remote_missing': {}}

@pytest.mark.trio
async def test_02_cmd(autojump_clock):
    async with stdtest(args={'init':123}) as st:
        s, = st.s
        async with st.client() as c:
            assert (await c.request("get_value", path=())).value == 123

            r = await run("client","-p",s.port,"set","-v","hello","foo")
            r = await run("client","-p",s.port,"set","-ev","'baz'","foo","bar")

            r = await run("client","-p",s.port,"get")
            assert r.stdout == "123\n"

            r = await run("client","-p",s.port,"get", "foo")
            assert r.stdout == "'hello'\n"

            r = await run("client","-p",s.port,"get", "foo", "bar")
            assert r.stdout == "'baz'\n"

            r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            assert r == {'nodes': {'test_0': 3}, 'known': {'test_0': ((1, 4),)}, 'missing': {}, 'remote_missing': {}, 'seq': 2}

            assert (await c.request("get_value", node="test_0", tick=1)).value == 123
            assert (await c.request("get_value", node="test_0", tick=2)).value == "hello"
            assert (await c.request("get_value", node="test_0", tick=3)).value == "baz"

            r = await c.request("set_value", path=(), value=1234, nchain=3)
            assert r.prev==123
            assert r.chain.tick == 4

            # does not yet exist
            with pytest.raises(ServerError):
                await c.request("get_value", node="test_0", tick=8)
            # has been superseded
            with pytest.raises(ServerError):
                await c.request("get_value", node="test_0", tick=1)
            # works
            assert (await c.request("get_value", node="test_0", tick=4)).value == 1234

            r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            assert r == {'nodes': {'test_0': 4}, 'known': {'test_0': ((1, 5),)}, 'missing': {}, 'remote_missing': {}}

@pytest.mark.trio
async def test_03_three(autojump_clock):
    async with stdtest(test_1={'init':125}, n=2, tocks=30) as st:
        s,si = st.s
        async with st.client(1) as ci:
            assert (await ci.request("get_value", path=())).value == 125

            r = await ci.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            # Various stages of integrating test_0
            assert \
                r == {'nodes': {'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                r == {'nodes': {'test_0': None, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                r == {'nodes': {'test_0': None, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {'test_0': (1,)}, 'remote_missing': {'test_0': (1,)}} or \
                r == {'nodes': {'test_1': 1, 'test_0': None,}, 'known': {'test_0': (1,), 'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                r == {'nodes': {'test_0': 0, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                False


            # This waits for test_0 to be fully up and running.
            async with st.client(0) as c:

                # At this point ci shall be fully integrated, and test_1 shall know this (mostly).
                r = await ci.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
                del r['tock']
                del r['seq']
                assert \
                        r == {'nodes': {'test_0': 0, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} or \
                        r == {'nodes': {'test_0': None, 'test_1': 1}, 'known': {'test_1': (1,)}, 'missing': {}, 'remote_missing': {}} 

                assert (await c.request("get_value", path=())).value == 125

                r = await c.request("set_value", path=(), value=126, nchain=3)
                assert r.prev==125
                assert r.chain.tick == 1
                assert r.chain.node == 'test_0'
                assert r.chain.prev.tick == 1
                assert r.chain.prev.node == 'test_1'
                assert r.chain.prev.tick == 1

                # This verifies that the chain entry for the initial update is gone
                # and the initial change is no longer retrievable.
                # We need the latter to ensure that there are no memory leaks.
                await anyio.sleep(1)
                r = await ci.request("set_value", path=(), value=127, nchain=3)
                assert r.prev==126
                assert r.chain.tick == 2
                assert r.chain.node == 'test_1'
                assert r.chain.prev.tick == 1
                assert r.chain.prev.node == 'test_0'
                assert r.chain.prev.tick == 1
                assert r.chain.prev.prev is None

                with pytest.raises(ServerError):
                    await c.request("get_value", node="test_1", tick=1)
                with pytest.raises(ServerError):
                    await ci.request("get_value", node="test_1", tick=1)
                
                # Now test that the internal states match.
                await anyio.sleep(1)
                r = await c.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
                del r['tock']
                del r['seq']
                assert r == {'nodes': {'test_0': 1, 'test_1': 2}, 'known': {'test_0': (1,), 'test_1': ((1,3),)}, 'missing': {}, 'remote_missing': {}}

            r = await ci.request("get_state", nodes=True, known=True, missing=True, remote_missing=True)
            del r['tock']
            del r['seq']
            assert r == {'nodes': {'test_0': 1, 'test_1': 2}, 'known': {'test_0': (1,), 'test_1': ((1,3),)}, 'missing': {}, 'remote_missing': {}}




# dist-kv client

import anyio
from anyio.exceptions import ClosedResourceError
import outcome
import msgpack
import socket
from async_generator import asynccontextmanager
from aioserf.util import ValueEvent
from .util import attrdict

import logging
logger = logging.getLogger(__name__)

_packer = msgpack.Packer(strict_types=False, use_bin_type=True).pack

__all__ = ['NodData', 'ManyData', 'open_client', 'StreamReply']

class NoData(ValueError):
    """No reply arrived"""

class ManyData(ValueError):
    """More than one reply arrived"""

class ServerError(RuntimeError):
    """The server sent us an error"""

class ServerClosedError(ServerError):
    pass

class ServerConnectionError(ServerError):
    pass

@asynccontextmanager
async def open_client(host, port, init_timeout=5):
    client = Client(host, port)
    async with anyio.create_task_group() as tg:
        async with  client._connected(tg, init_timeout=init_timeout) as client:
            yield client

class _StreamRequest:
    """
    This class represents the core of a multi-message reply.
    """
    result = None

    def __init__(self, client, seq):
        self.client = client
        self.seq = seq

    async def __call__(self, **params):
        params['seq'] = self.seq
        #logger.debug("Send %s", params)
        await self._socket.send_all(_packer(params))

class StreamReply:
    """
    This class represents a multi-message reply.
    """
    send_stop = True
    end_msg = None

    def __init__(self, conn, seq):
        self._conn = conn
        self.seq = seq
        self.q = anyio.create_queue(10000)

    async def set(self, res):
        if 'error' in res:
            await self.q.put(outcome.Error(ServerError(res.error)))
            return
        state = res.get('state', None)
        if state == 'end':
            await self.q.put(None)
            self.q = None
            self.end_msg = res
            return
        elif state is not None:
            logger.warning("Unknown state: %s", repr(state))
        del res['seq']
        await self.q.put(outcome.Value(res))
        return self

    def __iter__(self):
        raise RuntimeError("You need to use 'async for …'")
    __next__ = __iter__

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.q is None:
            raise StopAsyncIteration
        res = await self.q.get()
        if res is None:
            raise StopAsyncIteration
        return res.unwrap()

    async def cancel(self):
        if self.q is not None:
            await self.q.put(None)



class _SingleReply:
    """
    This class represents a single-message reply.
    It will delegate itself to a StreamReply if a multi message reply
    arrives.
    """
    send_stop = True

    def __init__(self, conn, seq):
        self._conn = conn
        self.seq = seq
        self.q = ValueEvent()

    async def set(self, res):
        if res.get('state') == 'start':
            res = StreamReply(self._conn, self.seq)
            await self.q.set(res)
            return res
        else:
            if 'error' in res:
                await self.q.set_error(ServerError(res.error))
            else:
                await self.q.set(res)

    def get(self):
        return self.q.get()
    
    async def cancel(self):
        pass


class Client:
    _server_init = None  # Server greeting

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self._seq = 0
        self._handlers = {}

    async def _handle_msg(self, msg):
        try:
            seq = msg.seq
        except KeyError:
            raise RuntimeError("Reader got out of sync: " + str(msg))
        try:
            hdl = self._handlers.pop(seq)
        except KeyError:
            logger.warn("Spurious message %s: %s", seq, msg)
            return

        res = await hdl.set(msg)
        if res:
            self._handlers[seq] = res


    async def _reader(self):
        """Main loop for reading
        """
        unpacker = msgpack.Unpacker(object_pairs_hook=attrdict, raw=False, use_list=False)

        async with anyio.open_cancel_scope(shield=True) as s:
            try:
                while True:
                    for msg in unpacker:
                        #logger.debug("Recv %s", msg)
                        await self._handle_msg(msg)

                    if self._socket is None:
                        break
                    try:
                        buf = await self._socket.receive_some(4096)
                    except ClosedResourceError:
                        return  # closed by us
                    if len(buf) == 0:  # Connection was closed.
                        raise ServerClosedError("Connection closed by peer")
                    unpacker.feed(buf)

            finally:
                hdl, self._handlers = self._handlers, None
                async with anyio.open_cancel_scope(shield=True):
                    for m in hdl.values():
                        await m.cancel()

    async def request(self, action, iter=None, seq=None, **params):
        """Send a request. Wait for a reply.

        Args:
          ``action``: what to do. If ``seq`` is set, this is the stream's
                      state, which should be ``None`` or ``'end'``.
          ``seq``: Sequence number to use. Only when terminating a
                   multi-message request.
          ``params``: whatever other data the action needs
          ``iter``: A flag how to treat multi-line replies.
                    ``True``: always return an iterator
                    ``False``: Never return an iterator, raise an error
                               if no or more than on reply arrives
                    Default: ``None``: return a StreamReply if multi-line
                                       otherwise return directly
        """
        if seq is None:
            act = "action"
            self._seq += 1
            seq = self._seq
        else:
            act = "state"

        if action is not None:
            params[act] = action
        params['seq'] = seq
        res = _SingleReply(self, seq)
        self._handlers[seq] = res

        #logger.debug("Send %s", params)
        await self._socket.send_all(_packer(params))
        res = await res.get()
        if iter is True and not isinstance(res, StreamReply):
            async def send_one(res):
                yield res
            res = send_one(res)
        elif iter is False and isinstance(res, StreamReply):
            rr = None
            async for r in res:
                if rr is not None:
                    raise MoreData(action)
            if rr is None:
                raise NoData(action)
        return res


    @asynccontextmanager
    async def stream(self, action, iter=None, **params):
        """Send a multi-message request. Wait for a reply at the end.

        Args:
          ``action``: what to do
          ``params``: whatever other data the action needs
          ``iter``: A flag how to treat multi-line replies.
                    ``True``: always return an iterator
                    ``False``: Never return an iterator, raise an error
                               if no or more than on reply arrives
                    Default: ``None``: return a StreamReply if multi-line
                                       otherwise return directly

        This is a context manager. Use it like this::

            async with client.strem("update", path="private storage".split()) as sender:
                with MsgReader("/tmp/msgs.pack") as f:
                    for msg in f:
                        await sender(msg)
            print(sender.result)
            # any server-side exception will be raised

        """
        self._seq += 1
        seq = self._seq

        params['action'] = action
        params['seq'] = seq
        params['state'] = 'start'

        #logger.debug("Send %s", params)
        await self._socket.send_all(_packer(params))

        res = _StreamRequest(self, seq)
        try:
            yield res
        except BaseException as exc:
            await self.request("error", seq=seq, error=str(exc))
            raise
        else:
            res.result = await self.request("end", seq=seq, iter=iter)


    @asynccontextmanager
    async def _connected(self, tg, init_timeout=5):
        """
        This async context manager handles the actual TCP connection to
        the DistKV server.
        """
        hello = ValueEvent()
        self._handlers[0] = hello
        
        #logger.debug("Conn %s %s",self.host,self.port)
        async with await anyio.connect_tcp(self.host, self.port) as sock:
            #logger.debug("ConnDone %s %s",self.host,self.port)
            try:
                self.tg = tg
                self._socket = sock
                await self.tg.spawn(self._reader)
                async with anyio.fail_after(init_timeout):
                    self._server_init = await hello.get()
                yield self
            except socket.error as e:
                raise ServerConnectionError(self.host, self.port) from e
            finally:
                if self._socket is not None:
                    await self._socket.close()
                    self._socket = None
                self.tg.cancel_scope.cancel()
                self.tg = None




# DistKV's data model

from __future__ import annotations

import weakref
import anyio
from range_set import RangeSet

from typing import Optional, List, Any

from .util import attrdict

from logging import getLogger
logger = getLogger(__name__)

class NodeDataSkipped(Exception):
    def __init__(self, node):
        self.node = node
    def __repr__(self):
        return "<%s:%s>" % (self.__class__.__name__, self.node)

class Node:
    """Represents one DistKV participant.
    """
    name: str = None
    tick: int = None
    _known: RangeSet = None # I have these as valid data
    _reported: RangeSet = None # somebody else reported these missing data for this node
    entries: dict = None

    def __new__(cls, name, tick=None, cache=None, create=True):
        try:
            self = cache[name]
        except KeyError:
            if not create:
                raise
            self = object.__new__(cls)
            self.name = name
            self.tick = tick
            self._known = RangeSet()
            self._reported = RangeSet()
            self.entries = {}
            cache[name] = self
        else:
            if tick is not None:
                if self.tick is None or self.tick < tick:
                    self.tick = tick
        return self

    def __init__(self, name, tick=None, cache=None, create=True):
        return

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, Node):
            other = other.name
        return self.name == other

    def __getitem__(self, item):
        return self.entries[item]

    def __contains__(self, item):
        return item in self.entries

    def __repr__(self):
        return "<%s: %s @%s>" % (self.__class__.__name__, self.name, self.tick)

    def seen(self, tick, entry=None, local=False):
        """An event with this tick is in the entry's chain.

        Args:
          ``tick``: The event affecting the given entry.
          ``entry``: The entry affected by this event.
          ``local``: The message was not broadcast, thus do not assume that
                     other nodes saw this.

        """
        self._known.add(tick)
        if not local:
            self._reported.discard(tick)
        if entry is not None:
            self.entries[tick] = entry

    def supersede(self, tick):
        """The event with this tick is no longer in the referred entry's chain.
        This happens when an entry is updated.
        
        Args:
          ``tick``: The event that once affected the given entry.
        """
        self.entries.pop(tick, None)

    def reported_known(self, range, local=False):
        """Some node said that these entries may have been superseded.
        
        Args:
          ``range``: The RangeSet thus marked.
          ``local``: The message was not broadcast, thus do not assume that
                     other nodes saw this.
        """
        self._known += range
        if not local:
            self._reported -= range

    def reported_missing(self, range):
        self._reported += range

    @property
    def local_known(self):
        """Values I have seen"""
        return self._known

    @property
    def local_missing(self):
        """Values I have not seen"""
        assert self.tick
        r = RangeSet(((1,self.tick+1),))
        r -= self._known
        return r

    @property
    def remote_missing(self):
        """Values from this node which somebody else has not seen"""
        return self._reported

class NodeEvent:
    """Represents any event originating at a node.

    Args:
      ``node``: The node thus affected
      ``tick``: Counter, timestamp, whatever
      ``prev``: The previous event, if any

    """
    def __init__(self, node:Node, tick:int = None, prev: NodeEvent = None, check_dup=True):
        self.node = node
        if tick is None:
            tick = node.tick
        self.tick = tick
        if check_dup and tick is not None and tick > 0:
            node.seen(tick)
        self.prev = None
        if prev is not None:
            self.prev = prev

    def __len__(self):
        if self.prev is None:
            return 1
        return 1+len(self.prev)

    def __repr__(self):
        return "<%s:%s @%s %s>" % (self.__class__.__name__, self.node, '-' if self.tick is None else self.tick, len(self))

    def __eq__(self, other):
        if other is None:
            return False
        return self.node == other.node and self.tick == other.tick

    def equals(self, other):
        """Check whether these chains are equal. Used for ping comparisons.

        The last two items may be missing from either chain.
        """
        if other is None:
            return self.prev is None or len(self.prev) <= 1
        if self != other:
            return False
        if self.prev is None:
            return other.prev is None or len(other.prev) <= 2
        elif other.prev is None:
            return self.prev is None or len(self.prev) <= 2
        else:
            return self.prev.equals(other.prev)

    def __lt__(self, other):
        if other is None:
            return False
        if self == other:
            return False
        while self.node != other.node:
            other = other.prev
            if other is None:
                return False
        return self.tick <= other.tick

    def __gt__(self, other):
        if other is None:
            return True
        if self == other:
            return False
        while self.node != other.node:
            self = self.prev
            if self is None:
                return False
        return self.tick >= other.tick


    def __lte__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __gte__(self, other):
        return self.__eq__(other) or self.__gt__(other)

    def __contains__(self, node):
        return self.find(node) is not None

    def find(self, node):
        """Return the position of a node in the node chain."""
        res = 0
        while self is not None:
            if self.node == node:
                return res
            res += 1
            self = self.prev
        return None

    def filter(self, node, dropped=None):
        """Return an event chain without the given node.

        If the node is not in the chain, the result is not a copy.
        """
        if self.node == node:
            if dropped is not None:
                dropped(self)
            return self.prev
            # Invariant: a node can only be in the chain once
            # Thus we can stop filtering after we encounter it.
        if self.prev is None:
            return self
        prev = self.prev.filter(node, dropped=dropped)
        if prev is self.prev:
            # No change, so return unmodified
            return self
        return NodeEvent(node=self.node, tick=self.tick, prev=prev)
        
    def serialize(self, nchain=100) -> dict:
        if not nchain:
            raise RuntimeError("A chopped NodeEvent must not be set at all")
        res = attrdict(node = self.node.name)
        if self.tick is not None:
            res.tick = self.tick
        if self.prev is None:
            res.prev = None
        elif nchain > 1:
            res.prev = self.prev.serialize(nchain-1)
        return res

    @classmethod
    def deserialize(cls, msg, cache, check_dup=True):
        if msg is None:
            return None
        msg = msg.get('chain',msg)
        tick = msg.get('tick', None)
        self = cls(node=Node(msg['node'], tick=tick, cache=cache), tick=tick, check_dup=check_dup)
        if 'prev' in msg:
            self.prev = cls.deserialize(msg['prev'], cache=cache, check_dup=check_dup)
        return self

    def attach(self, prev: NodeEvent=None, dropped=None):
        """Copy this node, if necessary, and attach a filtered `prev` chain to it"""
        if prev is not None:
            prev = prev.filter(self.node, dropped=dropped)
        if self.prev is not None or prev is not None:
            self = NodeEvent(node=self.node, tick=self.tick, prev=prev)
        return self

class _NotGiven:
    pass

class UpdateEvent:
    """Represents an event which updates something.
    """
    def __init__(self, event: NodeEvent, entry: Entry, new_value, old_value=_NotGiven):
        self.event = event
        self.entry = entry
        self.new_value = new_value
        if old_value is not _NotGiven:
            self.old_value = old_value

    def __repr__(self):
        if self.entry.chain == self.event:
            res = ""
        else:
            res = repr(self.event)+": "
        return "<%s:%s%s: %s→%s>" % (
                self.__class__.__name__,
                res,
                repr(self.entry),
                '-' if not hasattr(self,'old_value') else '' if self.new_value == self.entry.data else repr(self.old_value),
                '' if self.new_value == self.entry.data else repr(self.old_value),
            )

    def serialize(self, chop_path=0, nchain=2, with_old=False):
        res = self.event.serialize(nchain=nchain)
        res.path = self.entry.path[chop_path:]
        if with_old:
            res.old_value = self.old_value
            res.new_value = self.new_value
        else:
            res.value = self.new_value
        return res

    @classmethod
    def deserialize(cls, root, msg, cache):
        if 'value' in msg:
            value = msg.value
        else:
            value = msg.new_value
        event = NodeEvent.deserialize(msg, cache=cache)
        entry = root.follow(*msg.path, create=True)

        return UpdateEvent(event, entry, value)

class Entry:
    """This class represents one key/value pair
    """
    _parent: Entry = None
    name: str = None
    _path: List[str] = None
    _root: Root = None
    _data: bytes = None
    chain: NodeEvent = None

    monitors = None

    def __init__(self, name: Str, parent: Entry):
        self.name = name
        self._sub = {}
        self.monitors = set()

        if parent is not None:
            parent._add_subnode(self)
            self._parent = weakref.ref(parent)

    def _add_subnode(self, child: Entry):
        self._sub[child.name] = child

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, Entry):
            other = other.name
        return self.name == other

    def keys(self):
        return self._sub.keys()

    def values(self):
        return self._sub.values()

    def items(self):
        return self._sub.items()

    def __len__(self):
        return len(self._sub)

    def __bool__(self):
        return self._data is not None or len(self._sub) > 0

    @property
    def path(self):
        if self._path is None:
            parent = self.parent
            if parent is None:
                self._path = []
            else:
                self._path = parent.path + [self.name]
        return self._path

    def follow(self, *path, create = True):
        for name in path:
            child = self._sub.get(name, None)
            if child is None:
                if not create:
                    raise KeyError(name)
                child = Entry(name, self)
            self = child
        return self

    def __getitem__(self, name):
        return self._sub[name]

    @property
    def root(self):
        root = self._root
        if root is not None:
            root = root()
            if root is None:
                raise RuntimeError("Root node is gone")
            return root

        parent = self.parent
        if parent is None:
            return self
        root = parent.root()
        self._root = weakref.ref(root)
        return root

    @property
    def parent(self):
        parent = self._parent
        if parent is None:
            return None
        parent = parent()
        if parent is None:
            raise RuntimeError("Parent node is gone")
        return parent

    def __repr__(self):
        try:
            res = "<%s:%s" % (self.__class__.__name__, self.path)
            if self.chain is not None:
                res += "@%s" % (repr(self.chain),)
            if self.data is not None:
                res += " =%s" % (repr(self.data),)
            res += ">"
        except Exception as exc:
            res = "<%s:%s" % (self.__class__.__name__, str(exc))
        return res

    @property
    def data(self):
        return self._data

    async def set_data(self, event: NodeEvent, data: Any, local: bool = False, dropped=None):
        """This entry is updated by that event.

        Args:
          ``event``: The :cls:`NodeEvent` to base the update on.
          ``local``: Flag whether the event should be forwarded to watchers.

        Returns:
          The :cls:`UpdateEvent` that has been generated and applied.
        """
        event = event.attach(self.chain, dropped=dropped)
        evt = UpdateEvent(event, self, data, self._data)
        await self.apply(evt, local=local)
        return evt

    async def apply(self, evt:UpdateEvent, local: bool = False, dropped=None):
        """Apply this :cls`UpdateEvent` to me.
        
        Also, forward to watchers (unless ``local`` is set).
        """
        if evt.event == self.chain:
            assert self._data == evt.new_value, ("has:",self._data, "but should have:",evt.new_value)
            return
        if self.chain > evt.event: # already superseded
            return
        if not (self.chain < evt.event):
            logger.warn("*** inconsistency TODO ***")
            logger.warn("Node: %s", self)
            logger.warn("Current: %s", self.chain)
            logger.warn("New: %s", evt.event)
            return

        if not hasattr(evt, 'old_value'):
            evt.old_value = self._data
        if hasattr(evt, 'new_value'):
            self._data = evt.new_value
        else:
            self._data = evt.value
        if dropped is not None and self.chain is not None:
            dropped(evt.event, self.chain)
        self.chain = evt.event

        c = self.chain
        while c is not None:
            c.node.seen(c.tick, self)
            c = c.prev
        await self.updated(evt)

    async def walk(self, proc):
        """Call ``proc`` on this node and all its children)."""
        await proc(self)
        for v in list(self._sub.values()):
            await v.walk(proc)

    def serialize(self, chop_path=0, nchain=2):
        """Serialize this entry for msgpack.

        Args:
          ``chop_path``: If <0, do not return the entry's path.
                         Otherwise, do, but remove the first N entries.
          ``nchain``: how many change events to include.
        """
        res = attrdict(value=self._data)
        if self.chain is not None and nchain > 0:
            res.chain = self.chain.serialize(nchain=nchain)
        if chop_path >= 0:
            path = self.path
            if chop_path > 0:
                path = path[chop_path:]
            res.path = path
        return res

    async def updated(self, event: UpdateEvent):
        """Send an event to this node (and all its parents)'s watchers."""
        node = self
        while True:
            bad = set()
            for q in list(node.monitors):
                if q._distkv__free > 1:
                    q._distkv__free -= 1
                    await q.put(event)
                else:
                    bad.add(q)
            for q in bad:
                try:
                    if q._distkv__free > 0:
                        await q.put(None)
                    node.monitors.remove(q)
                except KeyError:
                    pass
                else:
                    await q.aclose()
            node = node.parent
            if node is None:
                break


    def log(self):
        """
        Returns an async context manager plus async iterator which reports
        all updates to this node.

        All concurrent loggers receive all updates.
        """
        return _RootLog(self)

    _counter = 0

    @property
    def counter(self):
        self._counter += 1
        return self._counter

class Watcher:
    """
    This helper class is used as an async context manager plus async
    iterator. It reports all updates to an entry (or its children).

    If a watcher terminates, sending to its channel has blocked.
    The receiver needs to take appropriate re-syncing action.
    """

    root: Entry = None
    q = None
    q_len = 100

    def __init__(self, root: Entry):
        self.root = root

    async def __aenter__(self):
        if self.q is not None:
            raise RuntimeError("You cannot enter this context more than once")
        self.q = anyio.create_queue(self.q_len)
        self.q._distkv__free = self.q_len
        self.root.monitors.add(self.q)
        return self

    async def __aexit__(self, *tb):
        self.root.monitors.remove(self.q)
        self.q = None

    def __aiter__(self):
        if self.q is None:
            raise RuntimeError("You need to enclose this with 'async with'")
        return self

    async def __anext__(self):
        if self.q is None:
            raise RuntimeError("Aborted. Queue filled?")
        res = await self.q.get()
        if res is None:
            raise RuntimeError("Aborted. Queue filled?")
        self.q._distkv__free += 1
        return res




from setuptools import setup, find_packages

LONG_DESC = open("README.rst", encoding="utf-8").read()

setup(
    name="distkv",
    use_scm_version={
        "version_scheme": "guess-next-dev",
        "local_scheme": "dirty-tag"
    },
    description="A distributed no-master key-value store",
    url="https://github.com/smurfix/distkv",
    long_description=LONG_DESC,
    author="Matthias Urlichs",
    author_email="matthias@urlichs.de",
    license="MIT -or- Apache License 2.0",
    packages=find_packages(),
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[
        "trio_click",
        "anyio",
        "range_set >= 0.2",
        "attrs",
        "aioserf >= 0.3.3",
    ],
    keywords=[
        "async", "key-values", "distributed",
    ],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Framework :: AsyncIO",
        "Framework :: Trio",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Database",
        "Topic :: Home Automation",
        "Topic :: System :: Distributed Computing",
    ],
    entry_points='''
    [console_scripts]
    distkv = distkv.command:cmd
    '''

)
